/*
** ymca.c
**
** Made by mycroft (mycroft)
** <mycroft@minithins.net> on virgaria.minithins.net
**
** Started on  Sun Dec 22 19:09:21 2002 mycroft
** Last update Tue Dec 24 22:02:50 2002 mycroft
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>

#include "ymca.h"

ymca create_socket_list (void)
{
  return (NULL);
}

int vider_struct_ymca_t (ymca elt)
{
  if (elt->fd)
    close (elt->fd);

  free (elt);

  return (0);
}

int destroy_socket_list (ymca *socket_liste)
{
  ymca tmp_ptr = *socket_liste;
  ymca tmp_ptr_next;

  while (tmp_ptr)
  {
    tmp_ptr_next = tmp_ptr->next;

    vider_struct_ymca_t (tmp_ptr);
    /* warning: This doesn't free ->data */

    tmp_ptr = tmp_ptr_next;
  }

  *socket_liste = NULL;

  return (0);
}

int is_fd_in_socket_list (ymca socket_liste, int fd)
{
  ymca tmp_ptr = socket_liste;

  while (tmp_ptr)
  {
    if (tmp_ptr->fd == fd)
     return (1);
    tmp_ptr = tmp_ptr->next;
  }

  return (0);
}

/* mode == 0, fd ==  1 => return static value */
/* mode == 0, fd == -1 => look for max */
/* mode > 0 => add fd */
/* mode < 0 => rem fd */
int get_max_fd_value_from_socket_list (ymca socket_liste, int mode, int fd)
{
  static int max = -1;
  ymca tmp_ptr = socket_liste;

  if(mode == 0 && fd == 1)
    return(max);
  else if(mode > 0 && max >= fd)
    return(max);
  else if(mode < 0 && max > fd)
    return(max);

  /* reset max */
  max = -1;

  while (tmp_ptr) 
  {
    if (max < tmp_ptr->fd)
      max = tmp_ptr->fd;
    tmp_ptr = tmp_ptr->next;
  }

  return (max);
}

ymca add_fd_to_socket_list (ymca *socket_liste, int fd)
{
  ymca new = NULL;

  if (!is_fd_in_socket_list (*socket_liste, fd))
  {
    if ((new = malloc (sizeof (struct ymca_t) * 1)))
    {
      new->next = *socket_liste;
      new->ptr_data = NULL;
      new->fd = fd;
      *socket_liste = new;

      /* rebuild max */
      get_max_fd_value_from_socket_list(*socket_liste, 1, fd);
    }
  }

  return (new);
}

ymca add_fd_data_to_socket_list (ymca *socket_liste, int fd, void *data)
{
  ymca new = NULL;

  if ((new = add_fd_to_socket_list (socket_liste, fd)))
    new->ptr_data = data;

  return (new);
}

void *get_data_from_socket_list_with_fd (ymca socket_liste, int fd)
{
  ymca tmp_ptr = socket_liste;
  void *data = NULL;

  while (tmp_ptr)
  {
    if (tmp_ptr->fd == fd)
    {
      data = tmp_ptr->ptr_data; 
      break;
    }

    tmp_ptr = tmp_ptr->ptr_data;
  }

  return (data);
}

ymca remove_fd_from_socket_list (ymca *socket_liste, int fd, void **data)
{
  ymca tmp_ptr = *socket_liste;
  ymca av_tmp_ptr = NULL;
  ymca n = NULL;

  while (tmp_ptr)
  {
    if (tmp_ptr->fd == fd)
    {
      n = tmp_ptr->next;

      if (data)
        *data = tmp_ptr->ptr_data;

      if (!av_tmp_ptr)
      {
        *socket_liste = tmp_ptr->next;
        vider_struct_ymca_t (tmp_ptr);
      }
      else
      {
        av_tmp_ptr->next = tmp_ptr->next;
        vider_struct_ymca_t (tmp_ptr);
      }

      /* rebuild if necessary the max fd */
      get_max_fd_value_from_socket_list(*socket_liste, -1, fd);

      break ;
    }

    av_tmp_ptr = tmp_ptr;
    tmp_ptr = tmp_ptr->next;
  }

  return (n);
}

int get_next_active_fd (ymca socket_liste, struct timeval *timeout)
{
  fd_set sfuck;
  static ymca ready_fd = NULL;
  ymca tmp_ptr;
  int f_ret = -1;
  int ret;

  if (!ready_fd)
  {
    FD_ZERO (&sfuck);

    tmp_ptr = socket_liste;
    while (tmp_ptr)
    {
      FD_SET (tmp_ptr->fd, &sfuck);
      tmp_ptr = tmp_ptr->next;
    }

    ret = select (get_max_fd_value_from_socket_list (socket_liste, 0, 1) + 1, 
                  &sfuck, NULL, NULL, timeout);

    switch (ret)
    {
      case -1:
            /* fix by Raphael POSS <raph@lrde.epita.fr> */
            /* if we were interrupted it could be because of SIGSTOP/SIGCONT,
             * in which case a timeout is very likely to have occured.
             * Reporting a timeout is therefore the safer bet. */
      if (errno == EINTR || errno == EAGAIN)
         return E_TIMEOUT;
      break;
      case 0:
        /* timeout */
        return(E_TIMEOUT);
      break;
      default:
        tmp_ptr = socket_liste;
        while (tmp_ptr)
        {
          if (FD_ISSET (tmp_ptr->fd, &sfuck))
            add_fd_to_socket_list (&ready_fd, tmp_ptr->fd);
          tmp_ptr = tmp_ptr->next;
        }
      break;
    }
  }

  if (ready_fd)
  {
    f_ret = ready_fd->fd;
    tmp_ptr = ready_fd->next;
    free (ready_fd);
    ready_fd = tmp_ptr;
  }

  return (f_ret);
}
